;;;; Copyright (C) 2016 Lily Carpenter

;;;; This program is free software: you can redistribute it and/or modify
;;;; it under the terms of the GNU Affero General Public License as published by
;;;; the Free Software Foundation, either version 3 of the License, or
;;;; (at your option) any later version.

;;;; This program is distributed in the hope that it will be useful,
;;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;;; GNU Affero General Public License for more details.

;;;; You should have received a copy of the GNU Affero General Public License
;;;; along with this program.  If not, see <http://www.gnu.org/licenses/>.

;;;; crypts-and-corpses.lisp

(in-package #:crypts-and-corpses)

(defstruct game-map
  (width 100 :type fixnum :read-only t)
  (height 40 :type fixnum :read-only t)
  (mobiles (make-hash-table) :type hash-table :read-only t))

(defclass entity ()
  ((width :initarg :width :type fixnum :initform 1)
   (height :initarg :height :type fixnum :initform 1)
   (location :initarg :location :initform 0 :type fixnum)))

(defclass tile-entity (entity)
  ((codepoint :initarg :codepoint :initform #x3f :type fixnum)))

(defclass viewport (entity)
  ((width :initarg :width :type fixnum :initform 100)
   (height :initarg :height :type fixnum :initform 40)
   (location :initarg :location :type fixnum :initform 0)))

(defclass mobile (tile-entity)
  ((health :initarg :health :type fixnum)
   (mana :initarg :mana :type fixnum)
   (health-regen :initarg :health-regen :initform 0 :type fixnum)
   (mana-regen :initarg :mana-regen :initform 0 :type fixnum)))

(defclass player-character (mobile)
  ((codepoint :initform #x263a)
   (ranged-attack-damage :initarg :ranged-attack-damage :initform 5 :type fixnum)
   (ranged-attack-cost :initarg :ranged-attack-cost :initform 5 :type fixnum)))

(defclass enemy (mobile) ())

(defclass minion (mobile) ())

(defun calc-coord (location map)
  (values (mod location (game-map-width map)) (floor (/ location (game-map-width map)))))

(defun get-window-size ()
  (apply 'values (mapcar #'parse-integer (split-sequence #\x (terminal-get "window.size" "0x0")))))

;; DEPRECATED until I find a use for it
;; (defun calc-location (x y map)
;;   (+ (* (game-map-width map) (cdr y)) (car x)))
;; (defgeneric set-location (object x y))
;; (defmethod set-location ((object tile-entity) x y)
;;   (setf (slot-value object 'location) (calc-location x y)))


(defgeneric end-of-turn (object))

(defmethod end-of-turn ((object mobile))
  (incf (slot-value object 'health) (slot-value object 'health-regen))
  (incf (slot-value object 'mana) (slot-value object 'mana-regen)))

(defgeneric move-left (object map))
(defgeneric move-right (object map))
(defgeneric move-up (object map))
(defgeneric move-down (object map))

(defgeneric update-location (object map old-location new-location))

(defun out-of-bounds-p (new-location map)
  (multiple-value-bind (x y) (calc-coord new-location map)
    (or (< x 0)
        (< y 0)
        (>= x (game-map-width map))
        (>= y (game-map-height map)))))

(defun overrun-p (old-location new-location map)
  (multiple-value-bind (old-x old-y) (calc-coord old-location map)
    (multiple-value-bind (new-x new-y) (calc-coord new-location map)
      (or (and (not (= old-y new-y))
               (not (= old-x new-x)))))))

(defun calculate-effective-location (object map location)
  (+
   location
   (1- (slot-value object 'width))
   (* (1- (slot-value object 'height)) (game-map-width map))))

(defun bad-move-p (object map old-location new-location)
  (let ((old-effective-location (calculate-effective-location object map old-location))
        (new-effective-location (calculate-effective-location object map new-location)))
    (or (out-of-bounds-p new-effective-location map)
        (out-of-bounds-p new-location map)
        (overrun-p old-effective-location new-effective-location map))))

(defmethod update-location ((object entity) map old-location new-location)
  (unless (bad-move-p object map old-location new-location)
    (setf (slot-value object 'location) new-location)))

(defmethod update-location ((object mobile) map old-location new-location)
  (unless (bad-move-p object map old-location new-location)
    (setf (slot-value object 'location) new-location)
    (setf (gethash new-location (game-map-mobiles map)) object)
    (remhash old-location (game-map-mobiles map))))

(defmethod move-left ((object entity) map)
  (let ((old-location (slot-value object 'location))
        (new-location (1- (slot-value object 'location))))
    (update-location object map old-location new-location)))

(defmethod move-up ((object entity) map)
  (let ((old-location (slot-value object 'location))
        (new-location (- (slot-value object 'location) (game-map-width map))))
    (update-location object map old-location new-location)))

(defmethod move-down ((object entity) map)
  (let ((old-location (slot-value object 'location))
        (new-location (+ (slot-value object 'location) (game-map-width map))))
    (update-location object map old-location new-location)))

(defmethod move-right ((object entity) map)
  (let ((old-location (slot-value object 'location))
        (new-location (1+ (slot-value object 'location))))
    (update-location object map old-location new-location)))

(defun intro-message ()
  (loop as key = (terminal-read) do
    (multiple-value-bind (win-width win-height) (get-window-size)
      (terminal-print (floor (- (/ win-width 2) 10)) (floor (1- (/ win-height 2))) "Crypts and Corpses")
      (terminal-print (floor (- (/ win-width 2) 11)) (floor (1+ (/ win-height 2))) "Press enter to start"))
    (terminal-refresh)
        until (= key bearlibterminal-ffi:+tk-enter+)))

(defun write-map-to-terminal (map viewport)
  (labels ((write-cell (key tile)
             (multiple-value-bind (x y) (calc-coord (- key (slot-value viewport 'location)) map)
               (terminal-put x y (slot-value tile 'codepoint)))))
    (maphash #'write-cell (game-map-mobiles map))))

(defun initialize-map (map player)
  (setf (gethash (slot-value player 'location) (game-map-mobiles map)) player)
  map)

(defun handle-input (key map viewport player)
  (cond
    ((= key bearlibterminal-ffi:+tk-up+) (move-up player map))
    ((= key bearlibterminal-ffi:+tk-down+) (move-down player map))
    ((= key bearlibterminal-ffi:+tk-right+) (move-right player map))
    ((= key bearlibterminal-ffi:+tk-left+) (move-left player map))
    ((= key bearlibterminal-ffi:+tk-w+) (move-up viewport map))
    ((= key bearlibterminal-ffi:+tk-a+) (move-left viewport map))
    ((= key bearlibterminal-ffi:+tk-d+) (move-right viewport map))
    ((= key bearlibterminal-ffi:+tk-s+) (move-down viewport map))
    ((= key bearlibterminal-ffi:+tk-space+) (update-location viewport map (slot-value viewport 'location) (slot-value player 'location)))))

(defun create-viewport (map)
  (multiple-value-bind (width height) (get-window-size)
    (when (> width (game-map-width map))
      (setf width (game-map-width map)))
    (when (> height (game-map-height map))
      (setf height (game-map-height map)))
    (make-instance 'viewport :width width :height height)))

(defun main-loop ()
  (let ((player (make-instance
                 'player-character

                 :location 0)))
    (loop with map = (initialize-map (make-instance 'game-map) player)
          with viewport = (create-viewport map)
          as key = (terminal-read) do
            (handle-input key map viewport player)
            (terminal-clear)
            (write-map-to-terminal map viewport)
            (terminal-refresh)
          until (or (= key bearlibterminal-ffi:+tk-close+)
                    (= key bearlibterminal-ffi:+tk-escape+)))))

(defun start ()
  (terminal-open)
  (terminal-set "window: size=100x40, title='Crypts and Corpses'")
  (terminal-set "font: ./fonts/DejaVuSansMono.ttf, size=16")
  (intro-message)
  (main-loop)
  (terminal-close))
